#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
DSuper8 project based on Joe Herman's rpi-film-capture.

Software modified by Manuel Ángel.

User interface redesigned by Manuel Ángel.

DS8Servidor.py: Server software main program.

Latest version: July 2021.
"""

from socket import (socket, AF_INET, SOCK_STREAM, SHUT_RDWR, SOL_SOCKET,
                    SO_REUSEADDR)

from threading import Event, Lock

from struct import pack

# Processes are used to control the stepper motor.
from multiprocessing import Queue, Event as Evento

from time import sleep

from logging import INFO, basicConfig, info, warning

# Our own modules.
from nbstreamreader import NonBlockingStreamReader

from streamer import ImageStreamer

from controlProcess import DS8Control, MotorDriver

from camera import DS8Camera

import codes as c

import config

class DS8Server():

    def __init__(self):

        self.framecount = 0

        # Works with other formats but large sizes make it very slow.
        # PNG image format has been tested. It turns out very slow.
        # self.img_format = "png"
        self.img_format = "jpeg"
        self.jpeg_quality = 100

        # This variable determines if we advance after each photo.
        # It is activated and deactivated by commands from the client.
        self.auto_advance = False

        # Exit event of the client order reading thread.
        self.readerExitEvent = Event()

        # Photo capture event.
        self.cap_event = Evento()

        # Application exit event.
        self.exit_event = Evento()

        # Queue for sending orders to the MotorDriver control process.
        self.motorQueue = Queue()

        # Camera control.
        config.cam = DS8Camera()

        # Motor and lighting control.
        self.control = DS8Control(self.cap_event, self.exit_event,
                                  self.motorQueue)

        # Motor turning process.
        self.driverprocess = MotorDriver(self.cap_event, self.exit_event,
                                         self.motorQueue)

        # Once the forward/backward signal of a frame is received, the client
        # is notified.
        self.driverprocess.frameMoveSig.connect(self.sendFrameMove)

        # Make connections.
        self.img_socket = socket(AF_INET, SOCK_STREAM)
        self.img_socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.img_socket.bind(('0.0.0.0', 8000))
        self.img_socket.listen(0)
        self.ctrl_socket = socket(AF_INET, SOCK_STREAM)
        self.ctrl_socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.ctrl_socket.bind(('0.0.0.0', 8001))
        self.ctrl_socket.listen(0)

        # Initiation of connections with the client program.
        self.setupConns(self.img_socket, self.ctrl_socket)

        # Blocking the connection used for sending images.
        self.connection_lock = Lock()

        # Connection command reading thread.
        self.ctrl_reader = NonBlockingStreamReader(config.ctrl_conn,
                                                   self.readerExitEvent)

        # We use separate threads for sending captured images.
        # 5 threads is attempted for extreme bracketing images support.
        config.pool = [ImageStreamer(self.connection_lock, config.pool_lock)
                       for i in range(5)]

        # The motor starts. According to tests carried out, it is advisable not
        # to activate and deactivate the motor continuously. Operation is more
        # precise and smoother if the motor is permanently active.
        self.control.motor_wake()

    # Make connections with the client program.
    def setupConns(self, img_socket, ctrl_socket):
        self.readerExitEvent.clear()
        info("Waiting for connection with the client...")
        config.img_conn = img_socket.accept()[0].makefile("wb")
        config.ctrl_conn = ctrl_socket.accept()[0].makefile("r")
        info("Client connection established")

    # Release connections to the client program.
    def releaseClient(self):
        self.readerExitEvent.set()
        config.img_conn.flush()
        config.img_conn.close()
        config.ctrl_conn.flush()
        config.ctrl_conn.close()

    # Execution of commands from the client program.

    def processCmd(self, cmdstr):

        cam = config.cam
        if not cmdstr:
            info("Empty command string")
            return 0
        cmdstr = cmdstr.replace("\n", "")
        info(cmdstr)
        cmd = cmdstr[0]
        setting = cmdstr[1:]

        # Settings defined in the user interface.

        # Initial settings.

        # zoomDial
        if cmd == c.SETZ:
            cam.setz(int(setting))

        # roiUpButton - roiDownButton
        elif cmd == c.SETY:
            cam.sety(int(setting))

        # roiLeftButton - roiRightButton
        elif cmd == c.SETX:
            cam.setx(int(setting))

        # lightCheckbox
        elif cmd == c.LIGHT_ON:
            self.control.light_on()
        elif cmd == c.LIGHT_OFF:
            self.control.light_off()

        # prevCheckBox
        elif cmd == c.PREVIEW_ON:
            self.control.light_on()
            self.sendLightOn()
            cam.startPreviewMode()
        elif cmd == c.PREVIEW_OFF:
            self.control.light_off()
            self.sendLightOff()
            cam.mode = cam.OFF

        # fRevButton
        elif cmd == c.MOTOR_FREV:
            self.control.motor_rev()

        # reverseButton
        elif cmd == c.MOTOR_REV:
            self.control.revFrame(1)

        # stopButton
        elif cmd == c.MOTOR_STOP:
            self.control.motor_stop()

        # forwardButton
        elif cmd == c.MOTOR_FWD:
            self.control.fwdFrame(1)

        # ffdButton
        elif cmd == c.MOTOR_FFD:
            self.control.motor_fwd()

        # Camera settings.

        # ISOcomboBox
        elif cmd == c.ISO:
            cam.iso = int(setting)

        # EV
        elif cmd == c.EXP_COMP:
            cam.exposure_compensation = int(setting)

        # awbBox
        # awb_mode = "off" allows the camera's white balance to be set manually
        # by adjusting the awb_gains parameter.
        elif cmd == c.AWB_MODE:
            cam.awb_mode = setting
            sleep(2)            
            if cam.awb_mode == "auto":
                self.sendGains()

        # blueGainBox
        elif cmd == c.GAIN_BLUE:
            cam.fixGains(1, float(setting) / 100)

        # redGainBox
        elif cmd == c.GAIN_RED:
            cam.fixGains(0, float(setting) / 100)

        # brightnessBox
        elif cmd == c.BRIGHTNESS:
            cam.brightness = int(setting)

        # contrastBox
        elif cmd == c.CONTRAST:
            cam.contrast = int(setting)

        # saturationBox
        elif cmd == c.SATURATION:
            cam.saturation = int(setting)

        # quitButton
        elif cmd == c.CLIENT_QUIT:
            cam.mode = cam.OFF
            config.exitEvent.set()
            # Time is allowed for the client to close connections.
            sleep(2)
            self.releaseClient()

        # Capture.

        # bracketing
        elif cmd == c.BRACKETING_SHOTS:
            cam.bracketing = int(setting)

        # stopsBox
        elif cmd == c.BRACKETING_STOPS:
            cam.stops = float(setting)

        # captureTestBtn
        elif cmd == c.TEST_PHOTO:
            self.auto_advance = False
            self.control.light_on()
            self.sendLightOn()
            cam.mode = cam.CAPTURING
            sleep(0.5)
            # Take and send a single photo.
            self.newImage()
            self.control.light_off()
            self.sendLightOff()

        # autoExpCheckBox
        elif cmd == c.AUTOEXP_ON:
            cam.autoExp = True

        elif cmd == c.AUTOEXP_OFF:
            cam.autoExp = False
            cam.ss = int(float(setting) * 1000)

        # timeExpBox
        elif cmd == c.FIX_EXPOSURE:
            cam.ss = int(float(setting) * 1000)

        # frameLcd
        elif cmd == c.UPDATE_FRAME:
            self.control.activateUpdateFrameSignals(True)

        elif cmd == c.NO_UPDATE_FRAME:
            self.control.activateUpdateFrameSignals(False)

        # activateMotorCheckBox
        elif cmd == c.ACTIVATE_MOTOR:
            self.control.motor_wake()

        elif cmd == c.DEACTIVATE_MOTOR:
            self.control.motor_sleep()

        # captureFrmRev10 - captureFrmRev - gotoCheckBox
        elif cmd == c.CAP_FRAME_REV:
            self.auto_advance = False
            num = int(setting) if setting else 1
            self.control.revFrame(num)

        # captureFrmAdv10 - captureFrmAdv - gotoCheckBox
        elif cmd == c.CAP_FRAME_ADV:
            self.auto_advance = False
            num = int(setting) if setting else 1
            self.control.fwdFrame(num)

        # captureStopBtn - capturePauseBtn pressed
        elif cmd == c.STOP_CAPTURE:
            self.auto_advance = False
            cam.mode = cam.OFF
            self.control.light_off()
            self.sendLightOff()

        # captureStartBtn - capturePauseBtn released
        elif cmd == c.START_CAPTURE:
            self.control.light_on()
            self.sendLightOn()
            cam.startCaptureMode()
            self.auto_advance = True
            self.newImage()

        # Advanced settings.

        # vflip
        elif cmd == c.VFLIP_ON:
            cam.vflip = True
        elif cmd == c.VFLIP_OFF:
            cam.vflip = False

        # hflip
        elif cmd == c.HFLIP_ON:
            cam.hflip = True
        elif cmd == c.HFLIP_OFF:
            cam.hflip = False

        # bw
        elif cmd == c.BW_ON:
            cam.color_effects = (128, 128)
        elif cmd == c.BW_OFF:
            cam.color_effects = None

        # drcBox
        elif cmd == c.DRC:
            cam.drc_strength = setting

        # sharpnessBox
        elif cmd == c.SHARPNESS:
            cam.sharpness = int(setting)

        # resolutionBox
        elif cmd == c.SET_SIZE:
            cam.setResize(int(setting))

        # stillCap
        elif cmd == c.VID_PORT_ON:
            cam.use_video_port = True
        elif cmd == c.VID_PORT_OFF:
            cam.use_video_port = False

        # New image requested from the client program.
        elif cmd == c.NEW_IMAGE:
            self.newImage()

    # Function to take and send photos at the request of the client program.
    def newImage(self):
        cam = config.cam

        # Exposure time is determined.
        if cam.autoExp:
            self.calcExpParmsAE()
            cam.ss = cam.exposure_speed

        self.calcFramerate(cam.ss)
        cam.shutter_speed = cam.ss

        # After the framerate parameter calculation, the autoexposure data is
        # sent.
        if cam.autoExp:
            self.sendSSAE()

        if cam.mode == cam.PREVIEWING:
            imgflag = "p"
            self.take_and_queue_photo(imgflag)
            # Manual exposure data is sent.
            if not cam.autoExp:
                self.sendSS()

            info("Taken preview image. Exposure time = " +
                 str(cam.exposure_speed) + " us - Framerate = " +
                 str(cam.framerate) + " fps - AG = " + str(cam.analog_gain) +
                 " - DG = " + str(cam.digital_gain))

        elif cam.mode == cam.CAPTURING:

            # This loop serves to ensure the stop of the motor and avoid
            # vibrations.
            while self.cap_event.is_set():
                sleep(0.1)

            for shot in range(1, cam.bracketing + 1):
                cam.shutter_speed = self.bracketSS(cam.stops, shot,
                                                   cam.bracketing, cam.ss)

                # Exposure data is sent.
                self.sendSS()

                # If we don't have free threads for the next photo (maybe due
                # to network congestion or a busy client) we wait until a
                # thread is available.
                # This should be weird, but it can happen.
                while not len(config.pool):
                    sleep(1)
                imgflag = ("s" if cam.bracketing == 1 else "a"
                           if shot < cam.bracketing else "b")

                self.take_and_queue_photo(imgflag)

                if imgflag == "s":
                    info("Single image taken. Exposure time = " +
                         str(cam.shutter_speed) + " us - Framerate = " +
                         str(cam.framerate) + " fps - AG = " +
                         str(round(float(cam.analog_gain), 2)) +
                         " - DG = " + str(round(float(cam.digital_gain), 2)))

                elif imgflag == "a":
                    info("Bracketing image taken. Exposure time = " +
                         str(cam.shutter_speed) + " us - Framerate = " +
                         str(cam.framerate) + " fps - AG = " +
                         str(round(float(cam.analog_gain), 2)) +
                         " - DG = " + str(round(float(cam.digital_gain), 2)))

                elif imgflag == "b":
                    info("Last bracketing image taken. Exposure time = " +
                         str(cam.shutter_speed) + " us - Framerate = " +
                         str(cam.framerate) + " fps - AG = " +
                         str(round(float(cam.analog_gain), 2)) +
                         " - DG = " + str(round(float(cam.digital_gain), 2)))

            if self.auto_advance:
                self.framecount += 1
                # Sending the signal to advance one frame.
                self.control.fwdFrame(1)

                info("Taken image " + str(self.framecount))

            else:
                info("Taken image")

    # Take and send a picture.

    def take_and_queue_photo(self, imgflag):
        cam = config.cam

        processor = None
        while not processor:
            with config.pool_lock:
                if config.pool:
                    processor = config.pool.pop()
            if processor:
                # Thread is available.
                processor.imgflag = imgflag.encode()

                # In preview or capture mode we use jpg format.
                # PNG image format has been tested. It turns out very slow.
                # cam.capture(processor.stream, self.img_format,
                #            use_video_port=cam.use_video_port)
                cam.capture(processor.stream, self.img_format,
                            quality=self.jpeg_quality,
                            use_video_port=cam.use_video_port)

                # Submit blue and red gains data if deemed necessary for
                # testing purposes.
                # self.sendGains()

                processor.exposureTime = cam.shutter_speed

                # Start the thread that sends the photo.
                processor.event.set()
            else:
                # If there are no processors, we wait until they are available.
                warning("There are no processors available. Capture on hold.")
                sleep(1)

    # Function for calculating automatic exposure parameters.
    def calcExpParmsAE(self):
        cam = config.cam

        # Setting shutter_speed = 0 automatically adjusts the exposure time.
        # The adjusted value can be read in the exposure_speed property of the
        # camera.
        cam.shutter_speed = 0

        # Fake images are taken to update the variable cam.exposure_speed.

        cam.capture(config.nullFile, self.img_format,
                    quality=self.jpeg_quality,
                    use_video_port=cam.use_video_port)

        prevAutoExp = cam.exposure_speed

        # This loop is used to stabilize the automatic exposure value.

        while True:
                cam.capture(config.nullFile, self.img_format,
                            quality=self.jpeg_quality,
                            use_video_port=cam.use_video_port)

                info("Current exposure = " + str(cam.exposure_speed) + " us")

                if prevAutoExp == cam.exposure_speed:
                    sleep(float(1 / cam.framerate))
                    cam.capture(config.nullFile, self.img_format,
                                quality=self.jpeg_quality,
                                use_video_port=cam.use_video_port)
                    info("Current exposure = " + str(cam.exposure_speed) + " us")
                    if prevAutoExp == cam.exposure_speed:
                        break
                    else:
                        prevAutoExp = cam.exposure_speed
                else:
                    prevAutoExp = cam.exposure_speed

    # This function calculates the exposure time for a certain number of
    # bracketing shots and a certain value of the step.
    def bracketSS(self, stops, shot, bkt, ss):
        cam = config.cam

        if bkt == 1:
            return ss
        else:
            # Provides a range of evenly spaced values between -1 and 1.
            # So if bkt = 1, is [-1,1] if 3, is [-1,0,1],
            # if 4 is [-1,-1/3,1/3,1], and so on.
            adj = (float(shot - 1) / float(bkt - 1) * 2) - 1

            shutter_speed = int(ss * 2**(adj * stops / 2))

            if shutter_speed > cam.maxExpTime:
                shutter_speed = cam.maxExpTime
            if shutter_speed < cam.minExpTime:
                shutter_speed = cam.minExpTime

            self.calcFramerate(shutter_speed)

            return shutter_speed

    # Sending the exposure time and analog and digital gains of the camera in
    # auto exposure.
    def sendSSAE(self, shutter_speed=True):
        cam = config.cam
        with self.connection_lock:
            # Flag e to point out that it is not an image, but autoexposure
            # data.
            config.img_conn.write("e".encode())
            config.img_conn.write(pack("<l", cam.exposure_speed))
            config.img_conn.write(pack("<f", round(cam.analog_gain, 2)))
            config.img_conn.write(pack("<f", round(cam.digital_gain, 2)))
            config.img_conn.write(pack('<f', round(cam.framerate, 1)))
            config.img_conn.flush()

    # Sending the exposure time and analog and digital gains of the camera.
    def sendSS(self):
        cam = config.cam
        with self.connection_lock:
            # Flag f to point out that it is not an image, but exposure data.
            config.img_conn.write("f".encode())
            config.img_conn.write(pack("<l", cam.shutter_speed))
            config.img_conn.write(pack("<f", round(cam.analog_gain, 2)))
            config.img_conn.write(pack("<f", round(cam.digital_gain, 2)))
            config.img_conn.write(pack('<f', round(cam.framerate, 1)))
            config.img_conn.flush()

    # Sending the gains of the blue and red.
    def sendGains(self):
        cam = config.cam
        with self.connection_lock:
            gred = int(cam.awb_gains[0] * 100)
            gblue = int(cam.awb_gains[1] * 100)
            info("Gains data sent to the client: blue = " + str(gblue) +
                 ", red = " + str(gred))
            # Flag g to point out that it is not an image, but color gain data.
            config.img_conn.write("g".encode())
            config.img_conn.write(pack('<i', gblue))
            config.img_conn.write(pack('<i', gred))
            config.img_conn.flush()

    # Sending forward or backward frame movement signal.
    def sendFrameMove(self, flag):

        with self.connection_lock:

            if flag == "c":
                config.img_conn.write("c".encode())
                info("Frame advance signal sent")
            elif flag == "C":
                config.img_conn.write("C".encode())
                info("Frame reverse signal sent")
            elif flag == "s":
                config.img_conn.write("m".encode())
                info("Motor stopped signal sent")

            config.img_conn.flush()

    # Sending notice light on.
    def sendLightOn(self):
        with self.connection_lock:
            config.img_conn.write("l".encode())
            config.img_conn.flush()

    # Sending notice light off.
    def sendLightOff(self):
        with self.connection_lock:
            config.img_conn.write("L".encode())
            config.img_conn.flush()

    # Calculation of the camera framerate parameter.
    def calcFramerate(self, ss):

        # According to documentation:
        # For V1 camera:
        # sensor mode = 1 resolución 1920x1080 px, framerate 1 - 30 fps
        # sensor mode = 2 resolución 2592x1944 px, framerate 1 - 15 fps
        #
        # Minimum framerate 1 fps and maximum framerate 15 fps are adopted for
        # this camera for all supported resolutions.
        #
        # For HQ camera:
        # sensor_mode = 1 resolution 2028x1080 px, framerate 0.1 - 50 fps.
        # sensor_mode = 2 resolution 2028x1520 px, framerate 0.1 - 50 fps.
        # sensor_mode = 3 resolution 4056x3040 px, framerate 0.005 - 10 fps.
        # sensor_mode = 4 resolution 1332x990 px, framerate 50.1 - 120 fps.
        #
        # 1332x990 px resolution is not considered relevant to this application
        # and has not been implemented.
        #
        # In tests carried out, it has been found that framerate values higher
        # than 30 fps cause problems, therefore framerate has been limited to
        # 30 fps.

        cam = config.cam

        framerate = 1

        if ss:
            framerate = round(1000000 / ss, 1)

        if config.cameraType == "V1":
            if framerate < 1:
                cam.framerate = 1

            elif framerate > 15:
                cam.framerate = 15

            else:
                cam.framerate = framerate

        elif config.cameraType == "HQ":

            if cam.resolution == cam.resolutions[0] and framerate < 0.1:
                cam.framerate = 0.1

            elif cam.resolution == cam.resolutions[0] and framerate > 30:
                cam.framerate = 30

            elif cam.resolution == cam.resolutions[1] and framerate < 0.1:
                cam.framerate = 0.1

            elif cam.resolution == cam.resolutions[1] and framerate > 30:
                cam.framerate = 30

            elif cam.resolution == cam.resolutions[2] and framerate < 0.005:
                cam.framerate = 0.005

            elif cam.resolution == cam.resolutions[2] and framerate > 10:
                cam.framerate = 10

            else:
                cam.framerate = framerate

    # Main loop of the server program.

    def run(self):
        try:

            while not config.exitEvent.is_set():
                data = self.ctrl_reader.readline(1)
                if data:
                    self.processCmd(data)
            if config.exitEvent.is_set():
                info("Exit event triggered")

        # Keyboard interrupt.
        except KeyboardInterrupt:
            info("Keyboard interrupt")
        except Exception as e:
            print(getattr(e, 'message', repr(e)))

        finally:
            info("Coming out...")
            # This is the thread event to stop the threads of previews and
            # image submission.
            config.exitEvent.set()
            # This is the process event to stop the motor.
            self.exit_event.set()
            # Terminate the image transmission threads that exist in the pool.
            while config.pool:
                streamer = config.pool.pop()
                streamer.terminated = True
                streamer.join()

            # Finish motor turning process.
            self.driverprocess.join()

            # Terminate order stream reading thread.
            self.ctrl_reader.thr.join()

            # Close connections.
            if config.img_conn:
                config.img_conn.close()
            if config.ctrl_conn:
                config.ctrl_conn.close()
            if self.img_socket:
                self.img_socket.shutdown(SHUT_RDWR)
                self.img_socket.close()
            if self.ctrl_socket:
                self.ctrl_socket.shutdown(SHUT_RDWR)
                self.ctrl_socket.close()

            info("Released client connections")

            # Stop the motor, turn off the light, release the GPIO port.
            self.control.cleanup()

            info("All connections closed")
            info("Finalized")

# Starting the server program.


if __name__ == "__main__":

    # Severity level of the log set to INFO.
    basicConfig(level=INFO, format="%(asctime)s - %(levelname)s - %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S")
    info("Starting")

    # Server DSuper8.
    server = DS8Server()

    # Starting the command stream reading thread.
    server.ctrl_reader.thr.start()

    # Starting the motor control process.
    server.driverprocess.start()

    # Server startup.
    server.run()
