# -*- coding: utf-8 -*-

"""
DSuper8 project based on Joe Herman's rpi-film-capture.

Software modified by Manuel Ángel.

User interface redesigned by Manuel Ángel.

camera.py: Camera configuration.

Latest version: July 2021.
"""

import picamera

from time import sleep

from logging import info

from config import cameraType

class DS8Camera(picamera.PiCamera):
    OFF = 0
    PREVIEWING = 1
    CAPTURING = 2

    if cameraType == "V1":
        resolutions = [(2592, 1944), (1600, 1200), (1440, 1080), (1296, 972),
                       (1024, 768), (800, 600)]
        res_preview = resolutions[0]
        res_capture = resolutions[0]

    elif cameraType == "HQ":
        resolutions = [(2028, 1080), (2028, 1520), (4056, 3040)]
        res_preview = resolutions[1]
        res_capture = resolutions[1]

    # Variable that distinguishes between the automatic exposure mode (True)
    # or the exposure time has been set manually (False)
    autoExp = True

    # Blue and red gains.
    bg = 0
    rg = 0

    def __init__(self):
        super(DS8Camera, self).__init__()
        self.setDefaults()
        info("Initialized camera")

    # Makes the initial camera settings.
    # They are set at server startup.
    # Once the connection is made with the client, it will send the settings
    # defined in the user interface.
    def setDefaults(self):
        self.mode = self.OFF

        # For V1 camera:
        # sensor mode = 1 -> resolution 1920x1080 px, framerate 1 - 30 fps
        # sensor mode = 2 -> resolution 2592x1944 px, framerate 1 - 15 fps
        # sensor mode = 3 -> resolution 2592x1944 px, framerate 0.1666 - 1 fps
        # sensor mode = 4 -> resolution 1296x972 px, framerate 1 - 42 fps
        # sensor mode = 5 -> resolution 1296x730 px, framerate 1 - 49 fps
        # sensor mode = 6 -> resolution 640x480 px, framerate 42.1 - 60 fps
        # sensor mode = 7 -> resolution 640x480 px, framerate 60.1 - 90 fps
        #
        # For HQ camera:
        # sensor_mode = 1 -> resolution 2028x1080 px, framerate 0,1 - 50 fps.
        # sensor_mode = 2 -> resolution 2028x1520 px, framerate 0,1 - 50 fps.
        # sensor_mode = 3 -> resolution 4056x3040 px, framerate 0,005 - 10 fps.
        # sensor_mode = 4 -> resolution 1332x990 px, framerate 50,1 - 120 fps.
        #
        # The 1332x990 px resolution is not considered of interest for this
        # application and has not been implemented.

        self.iso = 100

        # Exposure_mode auto is used for the camera to automatically select the
        # analog and digital gains, even if you set the exposure time manually.
        self.exposure_mode = "auto"
        self.awb_mode = "auto"
        self.bracketing = 1
        self.stops = 0

        # Variable to temporarily save the nominal exposure_speed value, to use
        # it in the calculation of bracketing exposures.
        self.ss = 2500

        # Maximum camera exposure time.
        # It is determined by the framerate variable.
        self.maxExpTime = 1000000

        # Minimum exposure time. It is fixed at 10 us.
        self.minExpTime = 10

        # The camera's still image port is slower than the video port but gives
        # higher quality images.
        # However, in tests carried out, the difference is negligible.
        self.use_video_port = True

        # Default resolution.
        self.resolution = self.res_capture

        self.sharpness = 0

        # Geometry of the region of interest to be captured by the camera.

        # Coordinates of the upper left corner.
        self.roiX = 0
        self.roiY = 0

        # Zoom value.
        # Determines the width and height of the region of interest.
        # Values between 0.4 and 1. They are given by the zoom control
        self.roiZ = 1

    # Initial settings.

    # zoomDial
    def setz(self, val):
        self.roiZ = float(val) / 1000
        self.roiX = (1 - self.roiZ) / 2
        self.roiY = self.roiX
        self.zoom = (self.roiX, self.roiY, self.roiZ, self.roiZ)

    # roiUpButton - roiDownButton
    def sety(self, val):
        self.roiY = float(val) / 1000
        self.zoom = (self.roiX, self.roiY, self.roiZ, self.roiZ)

    # roiLeftButton - roiRightButton
    def setx(self, val):
        self.roiX = float(val) / 1000
        self.zoom = (self.roiX, self.roiY, self.roiZ, self.roiZ)

    # prevCheckBox
    def startPreviewMode(self):
        # Resolution of the previous images if you want a particular resolution.
        self.resolution = self.res_preview
        self.mode = self.PREVIEWING
        info("Camera in preview mode")

    # Camera settings.

    # blueGainBox, redGainBox
    def fixGains(self, idx, value):
        gains = self.awb_gains
        self.awb_mode = "off"
        self.awb_gains = gains

        if (idx == 0):
            self.awb_gains = (value, self.awb_gains[1])

        if (idx == 1):
            self.awb_gains = (self.awb_gains[0], value)

        sleep(1)

        gblue = int(self.awb_gains[1] * 100)
        gred = int(self.awb_gains[0] * 100)

        info("Changed camera gains: blue = " + str(gblue)
             + ", red = " + str(gred))

    # Capture.

    # captureStartBtn

    def startCaptureMode(self):
        # Wait for the thread of the previous images to finish.
        sleep(2)
        self.resolution = self.res_capture
        sleep(1)
        self.mode = self.CAPTURING
        info("Camera in capture mode")

    # Advanced settings.

    # resolutionBox

    def setResize(self, num):
        self.resolution = self.resolutions[num]
        self.res_preview = self.resolutions[num]
        self.res_capture = self.resolutions[num]
        info("Sensor mode: " + str(self.sensor_mode) + " - Resolution: "
             + str(self.resolution))








