# -*- coding: utf-8 -*-

"""
DSuper8 project based on Joe Herman's rpi-film-capture.

Software modified by Manuel Ángel.

User interface redesigned by Manuel Ángel.

controlProcess.py: Control of lighting and motor movement.

Latest version: July 2021.
"""

from time import sleep

import RPi.GPIO as GPIO

from logging import info

from multiprocessing import Process, Event, Queue

from PyQt5.QtCore import pyqtSignal, QObject

# Our configuration module.
import config

# Management of the GPIO port for lighting control and motor movement.
class DS8Control():

    # Image capture event.
    cap_event = Event()

    # Application exit event.
    exit_event = Event()

    # Queue for sending orders to the motor control process MotorDriver.
    # The direction of rotation and the number of advance frames will be sent.
    motorQueue = Queue()

    def __init__(self, cap_event, exit_event, motorQueue):

        self.cap_event = cap_event
        self.exit_event = exit_event
        self.motorQueue = motorQueue

        # BCM pin numbering is used.
        # Illumination control: pin 0-> off 1-> on.
        self.light_pin = config.light_pin
        # Stepper motor control pins.
        # Direction: 0-> backward 1-> forward.
        self.dir_pin = config.dir_pin
        # Motor activation: 0-> enabled 1-> disabled.
        self.sleep_pin = config.sleep_pin
        # Spin pulses.
        self.pulse_pin = config.pulse_pin

        # They are used to activate and deactivate the motor.
        self.enabled = False
        self.disabled = True

        # Irrelevant messages are suppressed.
        GPIO.setwarnings(False)

        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self.light_pin, GPIO.OUT)
        GPIO.setup(self.dir_pin, GPIO.OUT)
        GPIO.setup(self.sleep_pin, GPIO.OUT)
        GPIO.setup(self.pulse_pin, GPIO.OUT)

        # Motor status:
        # 0 -> stop
        # 1 -> advance
        # 2 -> recoil
        self.motorstate = 0

    # lightCheckbox
    def light_on(self):
        GPIO.output(self.light_pin, True)

    def light_off(self):
        GPIO.output(self.light_pin, False)

    # fRevButton
    def motor_rev(self):
        if self.motorstate:
            self.motor_stop()
        # cb -> continuous recoil
        self.motorQueue.put(["cb", 0])
        self.cap_event.set()
        self.motorstate = -1

    # reverseButton
    def revFrame(self, num):
        if self.motorstate:
            self.motor_stop()
        while self.cap_event.is_set():
            sleep(0.5)
        # b -> recoil, num -> number of frames
        self.motorQueue.put(["b", num])
        self.cap_event.set()
        self.motorstate = -1

    # stopButton
    def motor_stop(self):
        # s -> motor stop
        if self.motorstate:
            self.motorQueue.put(["s", 0])
            self.motorstate = 0
        else:
            return

    # forwardButton
    def fwdFrame(self, num):
        if self.motorstate:
            self.motor_stop()
        while self.cap_event.is_set():
            sleep(0.5)
        # f -> advance, num -> number of frames
        self.motorQueue.put(["f", num])
        self.cap_event.set()
        self.motorstate = 1

    # ffdButton
    def motor_fwd(self):
        if self.motorstate:
            self.motor_stop()
        # cf -> continuous advance
        self.motorQueue.put(["cf", 0])
        self.cap_event.set()
        self.motorstate = 1

    # frameLcd

    # u -> enable sending of frame forward/backward signals.
    # U -> disable sending.
    def activateUpdateFrameSignals(self, update):
        if update:
            self.motorQueue.put(["u", 0])

        else:
            self.motorQueue.put(["U", 0])

    def cleanup(self):
        info("Cleaning the GPIO")
        self.light_off()
        self.motor_stop()
        GPIO.cleanup()

    def motor_wake(self):
        GPIO.output(self.sleep_pin, self.enabled)
        info("Motor on")
        sleep(0.5)

    def motor_sleep(self):
        GPIO.output(self.sleep_pin, self.disabled)
        info("Motor off")
        sleep(0.5)

# Very simple class designed to advance frames in another process during
# captures, so a different kernel can handle it and will not delay the
# photograph, or vice versa.
class MotorDriver(QObject, Process):

    # Photo capture event.
    cap_event = Event()

    # Application exit event.
    exit_event = Event()

    # Queue for receiving orders to the motor control process.
    # The direction of rotation and the number of advance frames will be
    # received.
    motorQueue = Queue()

    # One frame forward/backward notification signal.
    frameMoveSig = pyqtSignal(str)

    def __init__(self, cap_event, exit_event, motorQueue):
        super(MotorDriver, self).__init__()
        info("Starting MotorDriver")

        self.cap_event = cap_event
        self.exit_event = exit_event
        self.motorQueue = motorQueue

        # This variable determines the sending of forward and backward signals
        # of frames, for updating the client's position indicator.
        self.updateFrame = False

        # Stepper motor control pins.

        # Direction: 0->recoil 1->advance
        self.dir_pin = config.dir_pin

        # Motor activation: 0->on 1->off
        self.sleep_pin = config.sleep_pin

        # Spin pulses.
        self.pulse_pin = config.pulse_pin

        # Number of steps required to move forward/backward one frame.
        # For smooth operation we use 32 microsteps per step, has the
        # disadvantage of slowness.
        # Due to the wheel-pinion mesh, only half a revolution of the motor is
        # needed to advance one frame.
        self.steps_per_frame = config.steps_per_frame

        # Pulse width 25 microseconds.
        self.pw = 0.000025

        # They are used to determine the direction of rotation of the motor.
        self.backward = config.backward
        self.forward = config.forward

        # Variable that contains the order.
        self.order = ""

        # Variable that indicates continuous rotation of the motor if True.
        self.turn = False

        # Variable that determines the number of frames that the motor must
        # advance/reverse.
        self.numframes = 0

        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self.dir_pin, GPIO.OUT)
        GPIO.setup(self.sleep_pin, GPIO.OUT)
        GPIO.setup(self.pulse_pin, GPIO.OUT)

    # Main control loop of motor rotation.
    def run(self):
        info("Running motor turn process")
        try:
            while not self.exit_event.is_set():

                if self.cap_event.is_set() and not self.motorQueue.empty():

                    msg = self.motorQueue.get()
                    self.order = msg[0]
                    self.numframes = msg[1]

                else:
                    continue

                if self.order == "f":
                    GPIO.output(self.dir_pin, self.forward)
                    self.turnFrames(self.numframes)
                    # Motor stopped.
                    info("Motor stop")
                    self.cap_event.clear()
                    self.frameMoveSig.emit("s")

                elif self.order == "cf":
                    GPIO.output(self.dir_pin, self.forward)
                    self.turn = True
                    self.continuousTurn()
                    # Motor stopped.
                    info("Motor stop")
                    self.cap_event.clear()
                    self.frameMoveSig.emit("s")

                # In order to eliminate the small error that originates from
                # vertical scrolling when moving back frames, go back one
                # additional frame and then advance one frame.
                elif self.order == "cb":
                    GPIO.output(self.dir_pin, self.backward)
                    self.turn = True
                    self.continuousTurn()
                    sleep(0.5)
                    self.order = "f"
                    GPIO.output(self.dir_pin, self.forward)
                    self.turnFrames(1)
                    # Motor stopped.
                    info("Motor stop")
                    self.cap_event.clear()
                    self.frameMoveSig.emit("s")

                elif self.order == "b":
                    GPIO.output(self.dir_pin, self.backward)
                    self.turnFrames(self.numframes + 1)
                    sleep(0.5)
                    self.order = "f"
                    GPIO.output(self.dir_pin, self.forward)
                    self.turnFrames(1)
                    # Motor stopped.
                    info("Motor stop")
                    self.cap_event.clear()
                    self.frameMoveSig.emit("s")

                elif self.order == "s":
                    self.turn = False

                elif self.order == "u":
                    self.updateFrame = True

                elif self.order == "U":
                    self.updateFrame = False

                else:
                    continue

        except KeyboardInterrupt:
            info("Motor process terminated via keyboard")
        finally:
            info("End of the motor turning process")

    def turnFrames(self, numframes):
        if self.order == "f":
            if numframes == 1:
                info("1 frame advance")
            else:
                info(str(numframes) + " frames advance")

        if self.order == "b":
            if numframes == 1:
                info("1 frame reverse")
            else:
                info(str(numframes) + " frames reverse")

        for i in range(numframes):
            self.motorTurn()
            if self.order == "f" and self.updateFrame:
                self.frameMoveSig.emit("c")
            elif self.order == "b" and self.updateFrame:
                self.frameMoveSig.emit("C")

            # We check stop order.
            if not self.motorQueue.empty():
                msg = self.motorQueue.get()
                self.order = msg[0]
                if self.order == "s":
                    break

    def continuousTurn(self):
        if self.order == "cf":
            info("Continuous advance motor")
        elif self.order == "cb":
            info("Continuous reverse motor")

        # Continuous turning is done by full frames.
        while self.turn:
            self.motorTurn()
            if self.order == "cf" and self.updateFrame:
                self.frameMoveSig.emit("c")
            elif self.order == "cb" and self.updateFrame:
                self.frameMoveSig.emit("C")

            # We check stop order.
            if not self.motorQueue.empty():
                msg = self.motorQueue.get()
                self.order = msg[0]
                if self.order == "s":
                    self.turn = False
                else:
                    continue

    def motorTurn(self):
        for i in range(self.steps_per_frame):
            GPIO.output(self.pulse_pin, True)
            sleep(self.pw)
            GPIO.output(self.pulse_pin, False)
            sleep(self.pw)
